from urllib import request, parse
from urllib.error import HTTPError
from datetime import *
from hashlib import md5
from typing import List, Dict, Any, Optional
import json
import re
import os
import tempfile
from diesis import LyricsFinder, Logger, Config, TagHelper, Converter, Utils


class Song:
    __QUERY_FILTERS: List[str] = [r'\[[\S\s]+\]', r'\(radio\s+edit\)', r'^[0-9]+\.?']

    original_path: str = None
    path: str = None
    extension: str = None
    query: str = None
    minimal_query: str = None
    tags: Any = None
    tag_helper: TagHelper.TagHelper = None
    query_accuracy: int = 0
    title: str = None
    artist: str = None
    album_artist: str = None
    album: str = None
    cover_url: str = None
    cover_path: str = None
    date: str = None
    year: int = None
    disc_count: int = 0
    disc_number: int = 0
    track_count: int = 0
    track_number: int = 0
    genre: str = None
    group: str = None
    composer: str = None
    explicit: bool = False
    album_url: str = None
    track_url: str = None
    lyrics: str = None
    lyrics_writer: str = None
    found: bool = False

    def __load_tags(self) -> None:
        """
        Loads the song title and author embedded in file tags.
        :raise ValueError: If an unsupported file has been defined.
        """
        self.tags = TagHelper.TagHelper.generate_tag_object(self)
        self.tag_helper = TagHelper.TagHelper(self)
        self.tag_helper.fetch()

    @staticmethod
    def __filter_itunes_results(data: List[Dict[str, str]], query: str) -> int:
        """
        Analyzes results returned by iTunes and returns the index to the closest results according to song metadata.
        :param data: A list containing all the results returned by iTunes APIs.
        :type data: List[Dict[str, str]]
        :param query: A string containing the original query used to find song information through iTunes.
        :type query: str
        :return: An integer number representing the index of the closest entry found.
        :rtype: int
        """
        # Split the search query into words.
        words: List[str] = re.findall(r'[\w\']+', query.lower())
        max_score: int = 0
        min_length: int = 0
        index: int = 0
        i: int = 0
        query_ci: str = query.lower()
        # Check each result.
        for result in data:
            track_name_ci: str = result['trackName'].lower()
            artist_name_ci: str = result['artistName'].lower()
            if query_ci == track_name_ci:
                # This result matches the search query, use it.
                break
            if query_ci == (track_name_ci + ' ' + artist_name_ci) or query_ci == (artist_name_ci + ' ' + track_name_ci):
                # This result's track name and artist name values combined together matches the search query, use it.
                break
            # Get all the words contained within the title according to iTunes.
            title_words: List[str] = re.findall(r'[\w\']+', track_name_ci)
            title_length: int = len(title_words)
            # Find the common words between the words contained in the query and the words contained in this title.
            score: int = len(list(set(title_words) & set(words)))
            if score > max_score and (min_length == 0 or title_length < min_length):
                # Find the element having the highest number of matching words but the lowest length.
                index = i
                max_score = score
                min_length = title_length
            i += 1
        return index

    def __set_info_from_itunes(self, data: Any, query: str) -> None:
        """
        Loads information about this track from the most eligible result returned by the iTunes API.
        :param data: The whole response returned by iTunes.
        :type data: Any
        :param query: A string representing the search query used to get such information.
        :type query: str
        """
        if data['resultCount'] == 0:
            return
        Logger.Logger.log('Processing metadata from iTunes...')
        index: int = 0
        if data['resultCount'] != 1 and not Config.Config.get_strict_meta():
            Logger.Logger.log('Multiple results returned by iTunes, filtering them...')
            index = Song.__filter_itunes_results(data['results'], query)
        data = data['results'][index]
        # Add support for album artist and composer.
        self.title = data['trackName']
        self.artist = data['artistName']
        self.album_artist = data['artistName']
        self.genre = data['primaryGenreName']
        self.album = data['collectionName']
        release_date: datetime = datetime.strptime(data['releaseDate'], '%Y-%m-%dT%H:%M:%SZ')
        self.year = release_date.year
        self.cover_url = data['artworkUrl100'].replace('100x100bb.jpg', '1000x1000bb.jpg')
        self.disc_count = data['discCount']
        self.disc_number = data['discNumber']
        self.track_count = data['trackCount']
        self.track_number = data['trackNumber']
        # TODO: Currently iTunes doesn't support information about group and composer.
        # self.group = data['artistName']
        # self.composer = data['artistName']
        if data['trackExplicitness'] == 'explicit':
            self.explicit = True
        else:
            self.explicit = False
        self.album_url = data['artistViewUrl']
        self.track_url = data['trackViewUrl']

    def __get_filter_regex(self) -> str:
        """
        Returns the regex used to filter out unnecessary information from the simplified search query version.
        :return: A string containing the regular expression.
        :rtype: str
        """
        return '(' + ')|('.join(self.__QUERY_FILTERS) + ')'

    def __generate_search_query(self) -> None:
        """
        Generates the search query based on available information for this song.
        """
        if self.query_accuracy < 100:
            if self.title is not None and self.title != '' and self.artist is not None and self.artist != '':
                # Use the title and the artist name to find more information about the song.
                query: str = self.title + ' ' + self.artist
                query = re.sub(self.__get_filter_regex(), '', query)
                self.query = query
                # Remove unnecessary information in order to get a simpler query version.
                self.minimal_query = re.sub(r'\([\s\S]+\)', '', query).strip()
                self.query_accuracy = 100
                return
        if self.query_accuracy < 50:
            # No title nor artist name available, use the filename as search query.
            filename: str = os.path.basename(self.original_path)
            filename = os.path.splitext(filename)[0]
            query: str = filename.lower()
            query = re.sub(self.__get_filter_regex(), '', query)
            query = query.replace('_', ' ')
            query = query.strip()
            self.query = query
            self.minimal_query = re.sub(r'\([\s\S]+\)', '', query).strip()
            self.query_accuracy = 50

    def __init__(self, path: str, original_path: Optional[str] = None):
        """
        The class constructor.
        :param path: A string containing the path to the song file.
        :type path: str
        :param original_path: A string containing the original path, if the file has been moved.
        :type original_path: Optional[str]
        """
        if path:
            self.set_path(path, original_path)

    def get_tag_object(self) -> Any:
        """
        Returns the object that manages the song's tags.
        :return: An instance of the class that changes according to the audio file type.
        :rtype: Any
        """
        return self.tags

    def set_path(self, path: str, original_path: Optional[str] = None) -> None:
        """
        Sets the path to the song file.
        :param path: A string containing the path to the song file.
        :type path: str
        :param original_path: A string containing the original path, if the file has been moved.
        :type original_path: Optional[str]
        :raise ValueError: If an empty path is given.
        """
        if not path:
            raise ValueError('Path cannot be empty.')
        self.path = path
        if original_path:
            self.original_path = original_path
        else:
            self.original_path = path
        self.extension = os.path.splitext(path)[1].lower()[1:]
        # Reload tags according to new file.
        self.__load_tags()
        self.__generate_search_query()

    def get_path(self) -> Optional[str]:
        """
        Returns the path to the song file.
        :return: A string containing the path to the song file or "None" if no path has been defined.
        :rtype: Optional[str]
        """
        return self.path

    def get_original_path(self) -> Optional[str]:
        """
        Returns the path to the original file rather than its temporary copy.
        :return: A string containing the original path to the song file or "None" if no path has been defined.
        :rtype: Optional[str]
        """
        return self.original_path

    def get_extension(self) -> Optional[str]:
        """
        Returns the extension detected from the song filename.
        :return: A string containing the extension or "None" if no path has been defined.
        :rtype: Optional[str]
        """
        return self.extension

    def get_query_accuracy(self) -> int:
        """
        Returns the accuracy according to the search query that has been generated based on song's information.
        :return: An integer number representing the query accuracy from 0 (no query generated) to 100 (max accuracy).
        :rtype: int
        """
        return self.query_accuracy

    def set_title(self, title: str) -> None:
        """
        Sets the song title.
        :param title: A string containing the song title.
        :type title: str
        """
        self.title = title
        # Rebuild the song's search query to include the title defined.
        self.query_accuracy = 0
        self.__generate_search_query()

    def get_title(self) -> Optional[str]:
        """
        Returns the song title.
        :return: A string containing the song title, if no title has been defined nor found, "None" will be returned.
        :rtype: Optional[str]
        """
        return self.title

    def set_artist(self, artist: str) -> None:
        """
        Sets the song artist.
        :param artist: A string containing the artist, if no artist has been defined nor found, "None" will be returned.
        :type artist: str
        """
        self.artist = artist
        # Rebuild the song's search query to include the artist defined.
        self.query_accuracy = 0
        self.__generate_search_query()

    def get_artist(self) -> Optional[str]:
        """
        Returns the song artist.
        :return: A string containing the song artist, if no artist has been defined nor found, "None" will be returned.
        :rtype: Optional[str]
        """
        return self.artist

    def set_lyrics_writer(self, lyrics_writer: str) -> None:
        """
        Returns the author of the song's lyrics.
        :param lyrics_writer: A string containing the name(s) of the authors separated by a comma.
        :type lyrics_writer: str
        """
        self.lyrics_writer = lyrics_writer

    def get_lyrics_writer(self) -> Optional[str]:
        """
        Returns the author of the song's lyrics.
        :return: A string containing the name(s) of the authors, if no author has been found, will be returned None.
        :rtype: Optional[str]
        """
        return self.lyrics_writer

    def set_genre(self, genre: str) -> None:
        """
        Sets the song genre.
        :param genre: A string representing the song genre.
        :type genre: str
        """
        self.genre = genre

    def get_genre(self) -> Optional[str]:
        """
        Returns the song genre.
        :return: A string representing the song genre.
        :rtype: str
        """
        return self.genre

    def set_group(self, group: str) -> None:
        """
        Sets the name of the group who performed this song.
        :param group: A string containing the group name.
        :type group: str
        """
        self.group = group

    def get_group(self) -> Optional[str]:
        """
        Returns the name of the group who performed this song.
        :return: A string containing the group name.
        :rtype: str
        """
        return self.group

    def set_composer(self, composer: str) -> None:
        """
        Sets the name of the composer who wrote this song.
        :param composer: A string containing the composer name.
        :type composer: str
        """
        self.composer = composer

    def get_composer(self) -> Optional[str]:
        """
        Returns the name of the composer who wrote this song.
        :return: A string containing the composer name.
        :rtype: str
        """
        return self.composer

    def set_explicit(self, explicit: bool) -> None:
        """
        Marks this song as explicit or clear.
        :param explicit: If set to "True" the song will be marked as explicit, otherwise it will be considered clear.
        :type explicit: bool
        """
        self.explicit = explicit

    def get_explicit(self) -> bool:
        """
        Returns if the song has been marked as explicit or clear.
        :return: If the song has been marked as explicit will be returned "True", otherwise "False".
        :rtype: bool
        """
        return self.explicit

    def set_album_url(self, album_url: str) -> None:
        """
        Sets the URL to the album page on the iTunes website.
        :param album_url: A string containing the iTunes URL.
        :type album_url: str
        """
        self.album_url = album_url

    def get_album_url(self) -> Optional[str]:
        """
        Returns the URL to the album page on the iTunes website.
        :return: A string containing the iTunes URL.
        :rtype: str
        """
        return self.album_url

    def set_track_url(self, track_url: str) -> None:
        """
        Sets the URL to the song page on the iTunes store.
        :param track_url: A string containing the URL.
        :type track_url: str
        """
        self.track_url = track_url

    def get_track_url(self) -> Optional[str]:
        """
        Returns the URL to the song page on the iTunes store.
        :return: A string containing the URL.
        :rtype: str
        """
        return self.track_url

    def set_lyrics(self, lyrics: str) -> None:
        """
        Sets the song lyrics.
        :param lyrics: A string containing the full song lyrics.
        :type lyrics: str
        """
        self.lyrics = lyrics

    def get_lyrics(self) -> Optional[str]:
        """
        Returns the song lyrics.
        :return: A string containing the lyrics, if no lyrics were found, None will be returned instead.
        :rtype: Optional[str]
        """
        return self.lyrics

    def set_album_artist(self, album_artist: str) -> None:
        """
        Sets the artist of the whole album where this song is contained in.
        :param album_artist: A string containing the artist.
        :type album_artist: str
        """
        self.album_artist = album_artist

    def get_album_artist(self) -> Optional[str]:
        """
        Returns the artist of the whole album where this song is contained in.
        :return: A string containing the album artist, if no artist has been defined, "None" will be returned instead.
        :rtype: Optional[str]
        """
        return self.album_artist

    def set_album(self, album: str) -> None:
        """
        Sets the name of the album where this song is contained in.
        :param album: A string representing the album name.
        :type album: str
        """
        self.album = album

    def get_album(self) -> Optional[str]:
        """
        Returns the name of the album where this song is contained in.
        :return: A string containing the album name, if no album has been defined, "None" will be returned instead.
        :rtype: Optional[str]
        """
        return self.album

    def set_cover_path(self, path: str) -> None:
        """
        Sets the cover picture for this song.
        :param path: A string containing the path to the image file to use.
        :type path: str
        """
        self.cover_path = path
        # The cover URL is no longer required as a custom cover has been defined.
        self.cover_url = None

    def get_cover_path(self) -> Optional[str]:
        """
        Returns the path to the song cover image.
        :return: A string containing the path to the image, if no image has been defined, "None" will be returned.
        :rtype: Optional[str]
        """
        return self.cover_path

    def get_cover_url(self) -> Optional[str]:
        """
        Returns the URL of the image that has been found using the iTunes API.
        :return: A string containing the URL, if no image was found, "None" will be returned instead.
        :rtype: Optional[str]
        """
        return self.cover_url

    def set_year(self, year: int) -> None:
        """
        Sets the year when this song was made.
        :param year: An integer number greater than zero representing the year.
        :type year: int
        """
        if year <= 0:
            # Seriously, are you trying to convince me this song was made before Christ?
            self.year = None
            return
        self.year = year

    def get_year(self) -> Optional[int]:
        """
        Returns the year when this song was made.
        :return: An integer number representing the year, if no year has been defined, "None" will be returned.
        :rtype: Optional[int]
        """
        return self.year

    def set_disk(self, number: int, count: int) -> None:
        """
        Sets information about discs.
        :param number: An integer number greater than zero representing the number of the disc where this song is.
        :type number: int
        :param count: An integer number greater than zero representing the total number of discs.
        :type count: int
        """
        if count and count > 0:
            self.disc_count = count
        else:
            self.disc_count = None
        if number and number > 0:
            if self.disc_count and number > self.disc_count:
                # The disc number cannot be greater than the total disc number.
                self.disc_number = self.disc_count
            self.disc_number = number
            return
        self.disc_number = None

    def get_disc_count(self) -> Optional[int]:
        """
        Returns the total number of disc that compose the album where this song has been released in.
        :return: An integer number greater than zero representing the number or None if no information is available.
        :rtype: Optional[int]
        """
        return self.disc_count

    def get_disc_number(self) -> Optional[int]:
        """
        Returns the number of the disk where this song has been released in.
        :return: An integer number greater than zero representing the disc number or None if no number has been defined.
        :rtype: Optional[int]
        """
        return self.disc_number

    def set_track(self, number: int, count: int) -> None:
        """
        Sets information about tracks.
        :param number: An integer number greater than zero representing the track number of this song within the disc.
        :type number: int
        :param count: An integer number greater than zero representing the total number of tracks.
        :type count: int
        """
        if count and count > 0:
            self.track_count = count
        else:
            self.track_count = None
        if number and number > 0:
            if self.track_count and number > self.track_count:
                # The track number cannot be greater than the total track number.
                self.track_number = self.track_count
            self.track_number = number
            return
        self.track_number = None

    def get_track_count(self) -> Optional[int]:
        """
        Returns the track number of this song within the disc where this song is contained in.
        :return: An integer number greater than zero representing the track number or None if it is not available.
        :rtype: Optional[int]
        """
        return self.track_count

    def get_track_number(self) -> Optional[int]:
        """
        Returns the total count of tracks contained in the disk where this song has been released in.
        :return: An integer number greater than zero representing the number of track or None if it is not available.
        :rtype: Optional[int]
        """
        return self.track_number

    def is_found(self) -> bool:
        """
        Checks if online information for this song have been found or not.
        :return: If online information were found will be returned "True".
        :rtype: bool
        """
        return self.found

    def get_query(self, minimal: bool = False) -> Optional[str]:
        """
        Returns the search query generated for this song and used while looking up information through iTunes.
        :param minimal: If set to "True" the simpler search query version will be returned instead of the complete one.
        :type minimal: bool
        :return: A string containing the search query generated, if no query has been generated, "None" is returned.
        :rtype: Optional[str]
        """
        if minimal:
            return self.minimal_query
        return self.query

    @staticmethod
    def __fetch_from_url(url: str) -> Any:
        """
        Fetches the song information from a given iTunes URL.
        :param url: A string containing the URL where information must be fetched from.
        :type url: str
        :return: The JSON response or None if the response doesn't contain any result.
        :rtype: Any
        """
        song_information: Any = None
        try:
            # Send the request and load the returned contents.
            req = request.Request(url, headers={
                'User-Agent': Config.Config.get_user_agent()
            })
            response = request.urlopen(req)
            contents: str = response.read().decode('utf-8')
        except (HTTPError, TimeoutError) as ex:
            Logger.Logger.log_error(str(ex))
            Logger.Logger.log_error('Request failed for URL: ' + url)
            return
        # Parse the response from the endpoint as a JSON encoded string
        data: Any = json.loads(contents)
        # Check if response contains at least one result, otherwise return "None".
        if data['resultCount'] > 0:
            song_information = data
        return song_information

    def fetch_info(self, minimal: bool = False) -> None:
        """
        Fetch song information from the iTunes APIs based on the generated search query.
        :param minimal: If set to "True" the simpler search query version will be used instead of the complete one.
        :type minimal: bool
        :raise RuntimeError: If no song file has been defined.
        """
        self.found = False
        query: str = self.get_query(minimal)
        if not query:
            raise RuntimeError('No song has been defined.')
        # Prepare the API call.
        data: Any = None
        # Generate a list of english countries as alternatives to US to use whenever no result for a song is found.
        countries: List[str] = ['US', 'GB', 'AU']
        for country in countries:
            params: str = 'country=' + country + '&entity=song&limit=100&version=2&explicit=Yes&media=music'
            url: str = 'https://itunes.apple.com/search?term=' + parse.quote_plus(query) + '&' + params
            # Load results from iTunes API endpoint.
            data = Song.__fetch_from_url(url)
            if data:
                break
        if not data:
            title: str = Utils.Utils.str(self.title)
            Logger.Logger.log('Song ' + title + ' not found (query: ' + Utils.Utils.str(self.query) + ').')
            return
        # Update the song properties according to information returned by iTunes.
        self.__set_info_from_itunes(data, query)
        self.found = True
        if self.query_accuracy < 100:
            # If query accuracy was not the best, reprocess it using new information.
            self.__generate_search_query()

    def fetch_cover(self) -> None:
        """
        Fetches the cover image according to the URL returned by iTunes API.
        """
        self.cover_path = None
        if self.cover_url is None:
            Logger.Logger.log('No cover picture found for this song.')
            return
        Logger.Logger.log('Retrieving cover picture from iTunes...')
        url_hash: str = md5(self.cover_url.encode('utf-8')).hexdigest()
        filename: str = tempfile.gettempdir() + url_hash + '.jpg'
        try:
            request.urlretrieve(self.cover_url, filename)
            self.cover_path = filename
        except (HTTPError, TimeoutError) as ex:
            Logger.Logger.log_error(str(ex))
            Logger.Logger.log_error('Request failed for URL: ' + Utils.Utils.str(self.cover_url))
            self.cover_path = None

    def fetch_lyrics(self) -> None:
        """
        Fetches the song lyrics from the supported providers.
        """
        if self.artist is None or self.title is None:
            return
        Logger.Logger.log('Looking for song lyrics...')
        finder = LyricsFinder.LyricsFinder(self)
        finder.fetch()
        self.lyrics = finder.get_lyrics()
        self.lyrics_writer = finder.get_lyrics_writer()
        if not self.lyrics:
            Logger.Logger.log('No lyrics found for this song.')

    def get_all_info(self) -> None:
        """
        Fetches information about meta tags, cover picture and lyrics based on the song defined.
        """
        self.fetch_info(False)
        if not self.found and not Config.Config.get_strict_meta():
            Logger.Logger.log('No iTunes data found using full song name, retrying using a shorter version...')
            self.fetch_info(True)
        if not self.found:
            Logger.Logger.log('No available data for this song, skipping it...')
            return
        self.fetch_cover()
        self.fetch_lyrics()

    def convert(self, conversion_format: str) -> None:
        """
        Converts the song into the given format.
        :param conversion_format: A string containing the format the song must be converted into.
        :type conversion_format: str
        """
        if self.extension == conversion_format:
            return
        Logger.Logger.log('Converting the song into ' + conversion_format)
        converter: Converter.Converter = Converter.Converter(self)
        # Convert the file into the given format.
        new_path: str = converter.convert(conversion_format)
        if os.path.exists(self.path):
            os.remove(self.path)
        # Update the path and reload tags according to new file generated.
        self.set_path(new_path, self.original_path)

    def save(self) -> None:
        """
        Saves all the properties defined in this class instance to the file represented by this object.
        """
        if not self.found:
            return
        self.tag_helper.save()



from diesis.scrapers import LyricsScraper
from urllib import parse, request
from urllib.error import HTTPError
from http.client import RemoteDisconnected
from bs4 import BeautifulSoup
from diesis import Config, Logger
from typing import Tuple, Optional


class MusixMatch(LyricsScraper.LyricsScraper):
    def __search(self, minimal: bool = False) -> str:
        """
        Finds the URL where the song lyrics is available at based on the song's query string.
        :param minimal: If set to "True", a shorter version of the song query will be used instead of the complete one.
        :type minimal: bool
        :return: A string containing the URL found, if no URL is found, and empty string will be returned instead.
        :rtype: str
        :raise ValueError: If no search query has been defined for current song.
        """
        query: str = self.get_query(minimal)
        if not query:
            raise ValueError('No query defined for this song.')
        query = parse.quote(query)
        query = query.replace('%28', '(').replace('%29', ')')
        url: str = 'https://www.musixmatch.com/it/search/' + query
        try:
            # Send the request to the provider website.
            req = request.Request(url, headers={
                'User-Agent': Config.Config.get_user_agent()
            })
            response = request.urlopen(req)
            contents: str = response.read().decode('utf-8')
        except (HTTPError, RemoteDisconnected, TimeoutError) as ex:
            Logger.Logger.log_error(str(ex))
            Logger.Logger.log_error('Request failed for URL: ' + url)
            return ''
        # Parse the HTML page loaded.
        document = BeautifulSoup(contents, 'html.parser')
        main = document.select_one('div.main-panel')
        if main is None:
            return ''
        result = main.select_one('div.box-content')
        if result is None:
            return ''
        # Find the link to the lyrics page.
        link = main.select_one('a.title')
        if link is None:
            return ''
        # Returns the link as a text.
        return 'https://www.musixmatch.com' + link.get('href')

    @staticmethod
    def __load(url: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Extracts the song lyrics from the page that has been found in look up phase, then it will return it as a string.
        :param url: A string containing the URL where the song lyrics is located at.
        :type url: str
        :return: A tuple containing both the lyrics and its author(s) or both None if no lyrics is found.
        :rtype: Tuple[Optional[str], Optional[str]]
        :raise ValueError: If an empty URL is given.
        """
        if not str:
            raise ValueError('URL cannot be empty.')
        try:
            # Send the request to the page.
            req = request.Request(url, headers={
                'User-Agent': Config.Config.get_user_agent()
            })
            response = request.urlopen(req)
            # Load the HTML page contents.
            contents: str = response.read().decode('utf-8')
        except (HTTPError, RemoteDisconnected, TimeoutError) as ex:
            Logger.Logger.log_error(str(ex))
            Logger.Logger.log_error('Request failed for URL: ' + url)
            return None, None
        # Parse the HTML page.
        document = BeautifulSoup(contents, 'html.parser')
        lyrics: str = ''
        lyrics_writer: str = ''
        # Look up all the blocks containing the song lyrics.
        blocks = document.select('p.mxm-lyrics__content')
        if len(blocks) > 0:
            for block in blocks:
                # Concat the contents of every block found.
                inner = block.select_one('span')
                if inner is not None:
                    lyrics += '\n' + inner.getText().strip()
        lyrics = lyrics.strip()
        if lyrics is None:
            return None, None
        # Get the element that contains the lyrics writer.
        writer_block = document.select_one('p.mxm-lyrics__copyright')
        if writer_block is not None:
            # If the element containing the lyrics author is found, get its value and format the name(s) found in it.
            lyrics_writer = writer_block.getText().strip().lower().title()
            # Clean up the lyrics author string.
            if lyrics_writer.index('Writer(S): ') == 0:
                lyrics_writer = lyrics_writer[11:]
        return lyrics, lyrics_writer

    def fetch(self) -> None:
        """
        Searches and fetches the lyrics for the song defined.
        """
        lyrics: Tuple[Optional[str], Optional[str]] = (None, None)
        if self.query is not None:
            # If a search query for this song has been defined, search the lyrics using it.
            alternative: bool = False
            url: str = self.__search(False)
            if not url and self.minimal_query is not None and not Config.Config.get_strict_lyrics():
                # If no lyrics has been found using the normal search query, retry using a shorter and simpler one.
                Logger.Logger.log('No lyrics found using full song name, retrying using a shorter version...')
                alternative = True
                # Repeat the search telling the method to use the shorter query version.
                url = self.__search(True)
                if not url:
                    Logger.Logger.log('No lyrics found in anyway.')
            if url:
                lyrics = MusixMatch.__load(url)
                if not lyrics and not alternative and not Config.Config.get_strict_lyrics():
                    if self.minimal_query is not None:
                        # Lyrics were found but its page was empty, retry searching it using the shorter query version.
                        Logger.Logger.log('No lyrics found using full song name, retrying using a shorter version...')
                        # Repeat the search telling the method to use the shorter query version.
                        url = self.__search(True)
                        if url:
                            # Try again to fetch the song lyrics.
                            lyrics = MusixMatch.__load(url)
                        else:
                            Logger.Logger.log('No lyrics found in anyway.')
        self.lyrics = lyrics[0]
        self.lyrics_writer = lyrics[1]



from typing import Any


class Utils:
    @staticmethod
    def str(value: Any) -> str:
        """
        Converts a given variable into a string representation.
        :param value: The variable to be converted.
        :type value: Any
        :return: The string representation of the given variable.
        :rtype: str
        """
        if value is None:
            return ''
        return str(value)



from typing import Optional, Any, Set
from argparse import ArgumentParser
from diesis import Converter, FileScanner
import json
import os


class Config:
    WATERMARK: str = 'Processed by Diesis'
    USER_AGENT: str = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) ' \
                      'Chrome/35.0.1916.47 Safari/537.36 '

    source: Optional[str] = None
    destination: Optional[str] = None
    verbose: bool = False
    remove_original: bool = False
    watermark: bool = True
    rename: bool = False
    overwrite: bool = False
    strict_meta: bool = False
    strict_lyrics: bool = False
    format: Optional[str] = None
    bitrate: Optional[int] = None
    recursive: bool = False
    flatten: bool = False
    log_file: Optional[str] = None

    @staticmethod
    def __validate() -> None:
        """
        Validates configuration parameters loaded before running the application.
        """
        if not Config.source or not os.path.exists(Config.source):
            print('The given source file or directory does not exist, aborting.')
            quit()

    @staticmethod
    def get_watermark_text() -> str:
        """
        Returns the application signature that will be included in the comment tag in all the processed songs.
        :return: A string containing the application signature.
        :rtype: str
        """
        return Config.WATERMARK

    @staticmethod
    def get_user_agent() -> str:
        """
        Returns the user-agent string to use in HTTP requests.
        :return: A string containing the user-agent to use.
        :rtype: str
        """
        return Config.USER_AGENT

    @staticmethod
    def get_source_directory() -> Optional[str]:
        """
        Returns the path to the directory where the songs to process are located.
        :return: A string containing the path to the source directory
        :rtype: Optional[str]
        """
        return Config.source

    @staticmethod
    def get_destination_directory() -> Optional[str]:
        """
        Returns the path to the directory where the processed songs will be saved in.
        :return: A string containing the path to the destination directory, if no path is defined, "None" is returned.
        :rtype: Optional[str]
        """
        return Config.destination

    @staticmethod
    def is_verbose() -> bool:
        """
        Returns if the verbose logging has been turned on.
        :return: If the verbose logging is enabled will be returned "True".
        :rtype: bool
        """
        return Config.verbose

    @staticmethod
    def get_remove_original() -> bool:
        """
        Returns if the original file must be removed once processed.
        :return: If this option has been enabled will be returned "True".
        :rtype: bool
        """
        return Config.remove_original

    @staticmethod
    def get_watermark() -> bool:
        """
        Returns if a comment containing the application name must be included in the processed files.
        :return: If the watermark is going to be injected will be returned "True".
        :rtype: bool
        """
        return Config.watermark

    @staticmethod
    def get_rename() -> bool:
        """
        Returns if the file must be renamed using information that have been fetched from iTunes once processed.
        :return: If this option has been enabled will be returned "True".
        :rtype: bool
        """
        return Config.rename

    @staticmethod
    def get_overwrite() -> bool:
        """
        Returns if already existing files in destination directory can be overwritten by processed ones.
        :return: If older files can be overwritten will be returned "True".
        :rtype: bool
        """
        return Config.overwrite

    @staticmethod
    def get_strict_meta() -> bool:
        """
        Returns if the application can use a shorter version of the search query when looking for song info on iTunes.
        :return: If the short version of the song search query cannot be used will be returned "True".
        :rtype: bool
        """
        return Config.strict_meta

    @staticmethod
    def get_strict_lyrics() -> bool:
        """
        Returns if the application can use a shorter version of the search query when looking for the song lyrics.
        :return: If the short version of the song search query cannot be used will be returned "True".
        :rtype: bool
        """
        return Config.strict_lyrics

    @staticmethod
    def get_bitrate() -> Optional[int]:
        """
        Returns the bitrate to apply whenever converting an audio file into a lossy format.
        :return: An integer number representing the bitrate in kbps, if none has been defined, None will be returned.
        :rtype: Optional[int]
        """
        return Config.bitrate

    @staticmethod
    def get_format() -> Optional[str]:
        """
        Returns the format that the songs must be converted in once processed.
        :return: A string containing the format, if no format has been defined, "None" will be returned instead.
        :rtype: Optional[str]
        """
        return Config.format

    @staticmethod
    def get_recursive() -> bool:
        """
        Returns if songs contained in sub-directories must be processed as well.
        :return: If sub-directories will be processed will be returned "True"
        :rtype: bool
        """
        return Config.recursive

    @staticmethod
    def get_flatten() -> bool:
        """
        Returns if directory hierarchy found in source directory shouldn't be generated when using the recursive mode.
        :return: If source directory structure shouldn't be generated in destination folder will be returned "True".
        :rtype: bool
        """
        return Config.flatten

    @staticmethod
    def get_log_file() -> Optional[str]:
        """
        Returns the path to the file where log messages should be written in.
        :return: A string containing the path to the log file or None if no path has been defined.
        :rtype: Optional[str]
        """
        return Config.log_file

    @staticmethod
    def setup_from_cli() -> None:
        """
        Parses and loads the given CLI arguments.
        """
        parser: ArgumentParser = ArgumentParser(
            description='A simple auto-tagging tool for precise music collectors.'
        )
        # Set up the accepted arguments and options.
        parser.add_argument(
            'source_path',
            metavar='source',
            type=str,
            nargs='?',
            help='the source directory containing the music files to process.'
        )
        parser.add_argument(
            '--source',
            type=str,
            help='the source directory containing the music files to process.'
        )
        parser.add_argument(
            '--dest',
            nargs='?',
            type=str,
            help='the directory where the processed file must be stored in.'
        )
        parser.add_argument(
            '--format',
            nargs='?',
            type=str,
            help='the format that the songs files must be converted into once processed.'
        )
        parser.add_argument(
            '--bitrate',
            nargs='?',
            type=int,
            help='the bitrate to apply to converted files as kbps integer value.'
        )
        parser.add_argument(
            '--config',
            nargs='?',
            type=str,
            help='the path to a JSON configuration file where settings should be loaded from.'
        )
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='enable verbose logging for both normal messages and errors.'
        )
        parser.add_argument(
            '--remove_original',
            action='store_true',
            help='using this option, source files, once processed, are removed from source directory.'
        )
        parser.add_argument(
            '--nowatermark',
            action='store_true',
            help='prevents the application from adding a watermark comment within the processed files meta tags.'
        )
        parser.add_argument(
            '--rename',
            action='store_true',
            help='tells to the application to rename processed files according to information obtained from iTunes API.'
        )
        parser.add_argument(
            '--overwrite',
            action='store_true',
            help='allows the application to replace files having the same name in destination directory.'
        )
        parser.add_argument(
            '--strict_meta',
            action='store_true',
            help='prevent the application to use a simpler version of the search query to retrieve song information.'
        )
        parser.add_argument(
            '--strict_lyrics',
            action='store_true',
            help='prevent the application to use a simpler version of the search query to retrieve song lyrics.'
        )
        parser.add_argument(
            '--recursive',
            action='store_true',
            help='allows the application to load audio files in subdirectories as well as the given source directory.'
        )
        parser.add_argument(
            '--flatten',
            action='store_true',
            help='using this option, folder hierarchy found in source will not be reproduced in destination directory.'
        )
        parser.add_argument(
            '--log_file',
            nargs='?',
            type=str,
            help='the path to the log file where log messages should be written in.'
        )
        # GET the CLI arguments based on the registered values.
        args = parser.parse_args()
        if args.config:
            Config.load_from_json(args.config)
        # Set up the configuration class.
        source: Optional[str] = args.source
        if not source and args.source_path:
            source = args.source_path
        if source:
            Config.source = FileScanner.FileScanner.prepare_path(source)
        if args.dest:
            Config.destination = FileScanner.FileScanner.prepare_path(args.dest)
        if args.verbose is True:
            Config.verbose = True
        if args.remove_original is True:
            Config.remove_original = True
        if args.nowatermark is True:
            Config.watermark = False
        if args.rename is True:
            Config.rename = True
        if args.overwrite is True:
            Config.overwrite = True
        if args.strict_meta is True:
            Config.strict_meta = True
        if args.strict_lyrics is True:
            Config.strict_lyrics = True
        if args.recursive is True:
            Config.recursive = True
        if args.flatten is True:
            Config.flatten = True
        if args.format and args.format in Converter.Converter.get_supported_formats():
            # Check if the given format is supported by the conversion class.
            Config.format = args.format
        if args.bitrate and args.bitrate > 0:
            Config.bitrate = args.bitrate
        if args.log_file:
            Config.log_file = args.log_file
        # Validate all the loaded parameters before starting.
        Config.__validate()

    @staticmethod
    def load_from_json(path: str) -> None:
        """
        Loads the values of the properties of this class from a given JSON file.
        :param path: A string containing the path to the configuration file.
        :type path: str
        """
        if not path:
            raise ValueError('Invalid file path')
        # Open the given configuration file.
        with open(path, 'rb') as conf:
            contents: str = str(conf.read())
            # Parse its contents as JSON.
            data: Any = json.loads(contents)
            if 'source' in data and type(data['source']) is str and data['source']:
                Config.source = FileScanner.FileScanner.prepare_path(data['source'])
            if 'destination' in data and type(data['destination']) is str and data['destination']:
                Config.destination = FileScanner.FileScanner.prepare_path(data['destination'])
            if 'dest' in data and type(data['dest']) is str:
                Config.destination = data['dest']
            if 'verbose' in data and data['verbose'] is True:
                Config.verbose = True
            if 'remove_original' in data and data['remove_original'] is True:
                Config.remove_original = True
            if 'watermark' in data and data['watermark'] is False:
                Config.watermark = False
            if 'rename' in data and data['rename'] is True:
                Config.rename = True
            if 'overwrite' in data and data['overwrite'] is True:
                Config.overwrite = True
            if 'strict_meta' in data and data['strict_meta'] is True:
                Config.strict_meta = True
            if 'strict_lyrics' in data and data['strict_lyrics'] is True:
                Config.strict_lyrics = True
            formats: Set[str] = Converter.Converter.get_supported_formats()
            # Check if the given conversion format is supported.
            if 'format' in data and type(data['format']) is str and data['format'] and data['format'] in formats:
                Config.format = data['format']
            if 'bitrate' in data and type(data['bitrate']) is int and data['bitrate'] > 0:
                Config.bitrate = data['bitrate']
            if 'recursive' in data and data['recursive'] is True:
                Config.recursive = True
            if 'flatten' in data and data['flatten'] is True:
                Config.flatten = True



from typing import Optional, Tuple, Any
from diesis import Song, Logger
from diesis.scrapers import AZLyrics, MusixMatch


class LyricsFinder:
    lyrics: str = None
    lyrics_writer: str = None
    song: Song = None

    def __scrape(self, provider: int) -> Tuple[Optional[str], Optional[str]]:
        """
        Finds and fetches the song lyrics by scraping the given provider's website.
        :param provider: An integer number that allows to select the provider to scrape (1: Azlyrics, 2: MusixMatch).
        :type provider: int
        :return: A tuple containing both the lyrics and its author(s) or both None if no lyrics is found.
        :rtype: Tuple[Optional[str], Optional[str]]
        """
        scraper: Any = None
        if provider == 1:
            Logger.Logger.log('Querying "Azlyrics"...')
            scraper: AZLyrics.AZLyrics = AZLyrics.AZLyrics()
        if provider == 2:
            Logger.Logger.log('Querying "MusixMatch"...')
            scraper: MusixMatch.MusixMatch = MusixMatch.MusixMatch()
        if scraper is not None:
            scraper.set_query(self.song.get_query(False), self.song.get_query(True))
            scraper.fetch()
            return scraper.get_lyrics(), scraper.get_lyrics_writer()
        return None, None

    def set_song(self, song: Song) -> None:
        """
        Sets the song to process.
        :param song: An instance of the class "Song" representing the song.
        :type song: Song
        """
        self.song = song

    def get_song(self) -> Song:
        """
        Returns the song to process.
        :return: An instance of the class "Song" representing the song.
        :rtype: Song
        """
        return self.song

    def __init__(self, song: Song):
        """
        The class constructor.
        :param song: An instance of the class "Song" representing the song to process.
        :type song: Song
        """
        self.song = song

    def fetch(self) -> None:
        """
        Fetches the lyrics of the defined song by scraping all the available providers.
        :raise RuntimeError: If no song has been defined.
        """
        if self.song is None:
            raise RuntimeError('No song defined.')
        lyrics: Tuple[Optional[str], Optional[str]] = (None, None)
        provider: int = 1
        # Start fetching the lyrics from the first provider and go on to other ones until found.
        while not lyrics[0] and provider <= 2:
            lyrics = self.__scrape(provider)
            provider += 1
        if lyrics[0]:
            Logger.Logger.log('Song lyrics found and saved.')
        self.lyrics = lyrics[0]
        self.lyrics_writer = lyrics[1]

    def get_lyrics(self) -> Optional[str]:
        """
        Returns the lyrics found.
        :return: A string containing the lyrics found or None if no lyrics are found.
        :rtype: Optional[str]
        """
        return self.lyrics

    def get_lyrics_writer(self) -> Optional[str]:
        """
        Returns the authors who wrote the lyrics separated by a comma.
        :return: A string containing the lyrics authors, if no author is found, None will be returned instead.
        :rtype: Optional[str]
        """
        return self.lyrics_writer



from hashlib import md5
from shutil import copyfile
from datetime import datetime
from typing import Set, Optional, List
from pathlib import Path
from diesis import Logger, Song, Config, TagHelper, Converter, Utils
import tempfile
import os


class FileScanner:
    source: str = None
    destination: str = None

    @staticmethod
    def prepare_path(path: Optional[str]) -> Optional[str]:
        """
        Converts the given path into an absolute one.
        :param path: A string containing the path to prepare.
        :type path: Optional[str]
        :return: A string containing the processed path.
        :rtype: Optional[str]
        """
        if path and path[0] == '~':
            path = str(Path.home()) + path[1:]
        return path

    @staticmethod
    def get_allowed_file_types() -> Set[str]:
        """
        Returns all the supported formats according to the classes that are going to be involved in the process.
        :return: A set containing all the file extensions corresponding to the supported formats found.
        :rtype: Set[str]
        """
        if not Config.Config.get_format():
            # If file conversion is disabled, get only the formats supported by the "TagHelper" class.
            return TagHelper.TagHelper.get_supported_formats()
        return TagHelper.TagHelper.get_supported_formats() & Converter.Converter.get_supported_formats()

    def __move(self, song: Song.Song) -> None:
        """
        Renames the file corresponding to the given song using information fetched from iTunes as the new name.
        :param song: An object representing the song to rename.
        :type song: Song.Song
        """
        tmp_path: str = song.get_path()
        length: int = len(self.source) + 1
        original_base_path: str = song.get_original_path()[length:]
        # Use temporary file extension because it contains the converted file extension (in case of conversion).
        extension: str = os.path.splitext(tmp_path)[1].lower()
        filename: str = os.path.splitext(os.path.basename(original_base_path))[0]
        directory: str = ''
        if self.destination and original_base_path.find('/') and not Config.Config.get_flatten():
            # Recursive mode is enabled and original directory hierarchy must be reproduced in destination folder.
            directory = os.path.dirname(original_base_path) + '/'
        if Config.Config.get_rename():
            # New file must be renamed, building the name using information fetched from iTunes.
            filename = song.get_artist() + ' - ' + song.get_title()
        if self.destination:
            base_dir: str = self.destination + '/' + directory
        else:
            base_dir: str = self.source + '/'
        # Remove invalid characters from the user.
        filename = filename.replace('/', '-').replace('\\', '-')
        path: str = base_dir + filename + extension
        # Check if existing file overwrite is allowed or if the new file name doesn't exists.
        if Config.Config.get_overwrite() and os.path.exists(path):
            if directory and not os.path.exists(self.destination + '/' + directory):
                # The file is going to be moved, ensure the folder exists if directory hierarchy must be maintained.
                os.makedirs(self.destination + '/' + directory, 0o777, True)
            # Move temporary created file to its final destination folder.
            os.rename(tmp_path, path)
            return
        i: int = 1
        # Check if new file name exists, in this case, generate new names until a non-existing one is found.
        while os.path.exists(path):
            path = base_dir + filename + ' - ' + str(i) + extension
            i += 1
        if directory and not os.path.exists(self.destination + '/' + directory):
            os.makedirs(self.destination + '/' + directory, 0o777, True)
        # Move the file.
        os.rename(tmp_path, path)

    def __load_eligible_files(self, recursive: bool, context: Optional[str] = None) -> Set[str]:
        """
        Scans the source directory and returns all the supported audio files that will be processed.
        :param recursive: If set to "True" sub-directories will be scanned as well as the source one.
        :type recursive: bool
        :param context: A string containing the directory that must be scanned whenever running in recursive mode.
        :type context: Optional[str]
        :return: A set of strings containing the paths to the files to process.
        :rtype: Set[str]
        """
        file_list: Set[str] = set()
        if context:
            context = context + '/'
        else:
            context = ''
        # Get all the files contained within the given directory.
        files: List[str] = os.listdir(self.source + '/' + context)
        allowed_extensions: Set[str] = FileScanner.get_allowed_file_types()
        # Scan the directory.
        for file in files:
            if recursive and os.path.isdir(self.source + '/' + context + file):
                # If we are running in recursive mode, load the list of all the files contained in this directory.
                internal_list: Set[str] = self.__load_eligible_files(recursive, context + file)
                # Merge the list of all the file loaded from the sub-directory with main one.
                file_list = file_list.union(internal_list)
                continue
            extension: str = os.path.splitext(context + file)[1].lower()[1:]
            compliant: bool = False
            try:
                # Check if the file is supported according to its extension.
                if extension in allowed_extensions:
                    compliant = True
            except ValueError:
                continue
            if compliant:
                # This file appears to be supported, add it to the list.
                file_list.add(context + file)
        return file_list

    def __process_song(self, file: str) -> None:
        """
        Process a given file converting it into a song object.
        :param file: A string containing the path to the song file.
        :type file: str
        """
        Logger.Logger.log('Processing file: ' + file)
        # Generate a random and unique name fo the temporary file copy.
        time: int = datetime.now().microsecond
        extension: str = os.path.splitext(file)[1].lower()
        tmp_path: str = tempfile.gettempdir() + md5(file.encode('utf-8') + str(time).encode('utf-8')).hexdigest()
        tmp_path = tmp_path + extension
        # Create a copy of the original file where all edits will be made.
        copyfile(self.source + '/' + file, tmp_path)
        song: Song.Song = Song.Song(tmp_path, self.source + '/' + file)
        convert_format: Optional[str] = Config.Config.get_format()
        if convert_format:
            # Convert the song into the defined format before start precessing it.
            song.convert(convert_format)
        # Fetch song information from iTunes API.
        song.get_all_info()
        if song.is_found():
            # If information has been found save them.
            song.save()
            self.__move(song)
            if Config.Config.get_remove_original():
                try:
                    os.remove(self.source + '/' + file)
                except OSError:
                    pass
            Logger.Logger.log('Complete processing for file: ' + file + '\n')
            return
        try:
            os.remove(tmp_path)
        except OSError:
            pass
        Logger.Logger.log('Complete processing for file: ' + file + '\n')

    def __init__(self, directory: str = None):
        """
        The class constructor.
        :param directory: A string representing the path to the directory to scan where the songs' files are contained.
        :type directory: str
        """
        self.source = directory

    def set_source(self, source: str) -> None:
        """
        Sets the source directory where the file to process are stored in.
        :param source: A string representing the path to the source directory.
        :type source: str
        :raise ValueError: If an invalid path to the source directory is given.
        """
        if not source:
            raise ValueError('Invalid source directory path.')
        self.source = source

    def get_source(self) -> Optional[str]:
        """
        Returns the source directory where the file to process are stored in.
        :return: A string representing the path to the source directory or None if no source directory has been defined.
        :rtype: Optional[str]
        """
        return self.source

    def set_destination(self, destination: str):
        """
        Sets the directory where processed files will be stored in.
        :param destination: A string containing the path to the destination directory.
        :type destination: str
        """
        self.destination = destination
        return self

    def get_destination(self) -> Optional[str]:
        """
        Returns the path to the directory where processed files will be stored in.
        :return: A string representing the path to the destination directory or None if no directory has been defined.
        :rtype: Optional[str]
        """
        return self.destination

    def scan(self) -> None:
        """
        Processes all the eligible file found within the source directory that has been defined.
        :raise ValueError: If no source directory has been configured.
        """
        # Set source and destination directory, if no custom directories have been previously defined.
        if self.source is None:
            self.source = Config.Config.get_source_directory()
        if self.destination is None:
            self.destination = Config.Config.get_destination_directory()
        if self.source is None:
            raise ValueError('No source directory configured')
        if not os.path.isdir(self.source):
            # If a single file has been given instead of a whole directory, process it directly without looping.
            directory: str = os.path.dirname(self.source)
            filename: str = os.path.basename(self.source)
            # Sets the directory where this file is contained as source directory, then process it.
            self.source = directory
            self.__process_song(filename)
            return
        Logger.Logger.log('Loading files in ' + Utils.Utils.str(self.source))
        # Get the list of the files that are going to be processed.
        recursive: bool = Config.Config.get_recursive()
        file_list: Set[str] = self.__load_eligible_files(recursive)
        if not file_list:
            Logger.Logger.log('No eligible file found, exiting.')
            return
        if self.destination is not None:
            # Check if the destination directory exists, otherwise create it.
            if not os.path.exists(self.destination):
                os.mkdir(self.destination)
        Logger.Logger.log('Ready to process ' + str(len(file_list)) + ' files.')
        for file in file_list:
            self.__process_song(file)



from typing import Set, Any
from mutagen.mp4 import MP4, MP4Cover
from mutagen.id3 import ID3, TIT2, TPE1, TALB, TYER, TCON, USLT, TPOS, TRCK, APIC, COMM, TPE2, WOAF, PictureType, TEXT
from mutagen.flac import FLAC, Picture
from mutagen.aiff import AIFF
from diesis import Song, Config
import mutagen
import base64


class TagHelper:
    SUPPORTED_FORMATS: Set[str] = {'m4a', 'mp3', 'flac', 'aiff', 'aif', 'ogg'}

    song: Song = None

    @staticmethod
    def __str(value: Any) -> str:
        """
        Converts a given value into a string converting "None" values into an empty string instead of a literal "None".
        :param value: The value to convert.
        :type value: Any
        :return: The string representation of the given value.
        :rtype: str
        """
        if value is None:
            return ''
        return str(value)

    def __save_m4a(self) -> None:
        """
        Sets the file tags according to song properties using the format required by M4A files.
        """
        tags: Any = self.song.get_tag_object()
        # Set the tags value according to song properties.
        tags['©nam'] = [TagHelper.__str(self.song.get_title())]
        tags['©ART'] = [TagHelper.__str(self.song.get_artist())]
        tags['aART'] = [TagHelper.__str(self.song.get_album_artist())]
        tags['©alb'] = [TagHelper.__str(self.song.get_album())]
        tags['©day'] = [TagHelper.__str(self.song.get_year())]
        tags['©gen'] = [TagHelper.__str(self.song.get_genre())]
        tags['©wrt'] = [TagHelper.__str(self.song.get_composer())]
        tags['©grp'] = [TagHelper.__str(self.song.get_group())]
        if self.song.get_explicit():
            tags['rtng'] = [1]
        else:
            tags['rtng'] = [2]
        tags['©lyr'] = [TagHelper.__str(self.song.get_lyrics())]
        # TODO: Currently not supported song's properties: track_url, album_url, lyrics_writer
        i: int = self.song.get_disc_number()
        total: int = self.song.get_disc_count()
        if i > 0 and total > 0:
            tags['disk'] = [(
                self.song.get_disc_number(),
                self.song.get_disc_count()
            )]
        i = self.song.get_track_number()
        total = self.song.get_track_count()
        if i > 0 and total > 0:
            tags['trkn'] = [(
                self.song.get_track_number(),
                self.song.get_track_count()
            )]
        # Generate the tag representation of the cover image, if defined.
        if self.song.get_cover_path() is not None:
            with open(self.song.get_cover_path(), 'rb') as cover:
                tags['covr'] = [MP4Cover(cover.read(), imageformat=MP4Cover.FORMAT_JPEG)]
        if Config.Config.get_watermark():
            # Add the application watermark.
            tags['©cmt'] = Config.Config.get_watermark_text()
        # Save the edited tags to the file.
        tags.save()

    def __save_id3(self) -> None:
        """
        Sets the file tags according to song properties using the ID3 format.
        """
        tags: Any = self.song.get_tag_object()
        # Set the tags value according to song properties.
        tags['TIT2'] = TIT2(encoding=3, text=TagHelper.__str(self.song.get_title()))
        tags['TPE1'] = TPE1(encoding=3, text=TagHelper.__str(self.song.get_artist()))
        tags['TPE2'] = TPE2(encoding=3, text=TagHelper.__str(self.song.get_album_artist()))
        tags['TALB'] = TALB(encoding=3, text=TagHelper.__str(self.song.get_album()))
        tags['TYER'] = TYER(encoding=3, text=TagHelper.__str(self.song.get_year()))
        tags['TCON'] = TCON(encoding=3, text=TagHelper.__str(self.song.get_genre()))
        tags['TCOM'] = TCON(encoding=3, text=TagHelper.__str(self.song.get_composer()))
        tags['WOAF'] = WOAF(encoding=3, text=TagHelper.__str(self.song.get_track_url()))
        tags['USLT'] = USLT(encoding=3, text=TagHelper.__str(self.song.get_lyrics()))
        tags['TEXT'] = TEXT(encoding=3, text=TagHelper.__str(self.song.get_lyrics_writer()))
        # TODO: Currently not supported song's properties: explicit, album_url, group
        i: int = self.song.get_disc_number()
        total: int = self.song.get_disc_count()
        if i > 0 and total > 0:
            c: str = TagHelper.__str(self.song.get_disc_number()) + '/' + TagHelper.__str(self.song.get_disc_count())
            tags['TPOS'] = TPOS(encoding=3, text=c)
        i = self.song.get_track_number()
        total = self.song.get_track_count()
        if i > 0 and total > 0:
            c: str = TagHelper.__str(self.song.get_track_number()) + '/' + TagHelper.__str(self.song.get_track_count())
            tags['TRCK'] = TRCK(encoding=3, text=c)
        # Generate the tag representation of the cover image, if defined.
        if self.song.get_cover_path() is not None:
            with open(self.song.get_cover_path(), 'rb') as cover:
                tags['APIC'] = APIC(encoding=3, mime='image/jpeg', type=3, data=cover.read())
        if Config.Config.get_watermark():
            # Add the application watermark.
            tags['COMM'] = COMM(encoding=3, text=Config.Config.get_watermark_text())
        # Save the edited tags to the file.
        tags.save()

    def __save_flac(self) -> None:
        """
        Sets the file tags according to song properties using the format required by FLAC files.
        """
        tags: Any = self.song.get_tag_object()
        # Set the tags value according to song properties.
        tags['title'] = [TagHelper.__str(self.song.get_title())]
        tags['artist'] = [TagHelper.__str(self.song.get_artist())]
        tags['albumartist'] = [TagHelper.__str(self.song.get_album_artist())]
        tags['album'] = [TagHelper.__str(self.song.get_album())]
        tags['year'] = [TagHelper.__str(self.song.get_year())]
        tags['genre'] = [TagHelper.__str(self.song.get_genre())]
        tags['composer'] = [TagHelper.__str(self.song.get_composer())]
        tags['wwwaudiofile'] = [TagHelper.__str(self.song.get_track_url())]
        tags['wwwartist'] = [TagHelper.__str(self.song.get_album_url())]
        i: int = self.song.get_disc_number()
        total: int = self.song.get_disc_count()
        if i > 0 and total > 0:
            c: str = TagHelper.__str(self.song.get_disc_number()) + '/' + TagHelper.__str(self.song.get_disc_count())
            tags['discnumber'] = [c]
        i = self.song.get_track_number()
        total = self.song.get_track_count()
        if i > 0 and total > 0:
            c: str = TagHelper.__str(self.song.get_track_number()) + '/' + TagHelper.__str(self.song.get_track_count())
            tags['tracknumber'] = [c]
        tags['lyrics'] = [TagHelper.__str(self.song.get_lyrics())]
        tags['lyricist'] = [TagHelper.__str(self.song.get_lyrics_writer())]
        tags['grouping'] = [TagHelper.__str(self.song.get_group())]
        # TODO: Currently not supported song's properties: explicit
        if self.song.get_cover_path() is not None:
            with open(self.song.get_cover_path(), 'rb') as cover:
                # Generate the picture object representing the cover image.
                picture = Picture()
                picture.data = cover.read()
                picture.type = PictureType.COVER_FRONT
                picture.mime = u'image/jpeg'
                picture.width = 1000
                picture.height = 1000
                picture.depth = 16
                # Remove all the pictures from this file.
                tags.clear_pictures()
                # Add the picture that has been found.
                tags.add_picture(picture)
        if Config.Config.get_watermark():
            # Add the application watermark.
            tags['comment'] = [Config.Config.get_watermark_text()]
        # Save the edited tags to the file.
        tags.save()

    def __save_aiff(self) -> None:
        """
        Sets the file tags according to specification for AIFF audio format.
        """
        # I figured out AIFF files use ID3 tags as well as MP3 do.
        self.__save_id3()
        return

    def __save_ogg(self) -> None:
        """
        Sets the file tags according to the standard used in OGG files.
        """
        tags: Any = self.song.get_tag_object()
        # Set the tags value according to song properties.
        tags['title'] = [TagHelper.__str(self.song.get_title())]
        tags['artist'] = [TagHelper.__str(self.song.get_artist())]
        tags['albumartist'] = [TagHelper.__str(self.song.get_album_artist())]
        tags['album'] = [TagHelper.__str(self.song.get_album())]
        tags['year'] = [TagHelper.__str(self.song.get_year())]
        tags['genre'] = [TagHelper.__str(self.song.get_genre())]
        tags['composer'] = [TagHelper.__str(self.song.get_composer())]
        tags['wwwaudiofile'] = [TagHelper.__str(self.song.get_track_url())]
        tags['wwwartist'] = [TagHelper.__str(self.song.get_album_url())]
        i: int = self.song.get_disc_number()
        total: int = self.song.get_disc_count()
        if i > 0 and total > 0:
            c: str = TagHelper.__str(self.song.get_disc_number()) + '/' + TagHelper.__str(self.song.get_disc_count())
            tags['discnumber'] = [c]
        i = self.song.get_track_number()
        total = self.song.get_track_count()
        if i > 0 and total > 0:
            c: str = TagHelper.__str(self.song.get_track_number()) + '/' + TagHelper.__str(self.song.get_track_count())
            tags['tracknumber'] = [c]
        tags['lyrics'] = [TagHelper.__str(self.song.get_lyrics())]
        tags['lyricist'] = [TagHelper.__str(self.song.get_lyrics_writer())]
        tags['grouping'] = [TagHelper.__str(self.song.get_group())]
        if self.song.get_cover_path() is not None:
            with open(self.song.get_cover_path(), 'rb') as cover:
                # Encode the image as Base64 string.
                tags['METADATA_BLOCK_PICTURE'] = str(base64.b64encode(cover.read()))
        if Config.Config.get_watermark():
            # Add the application watermark.
            tags['comment'] = [Config.Config.get_watermark_text()]
        tags.save()

    @staticmethod
    def get_supported_formats() -> Set[str]:
        """
        Returns the extensions of all the supported file types.
        :return: A set containing the supported file extensions.
        :rtype: Set[str]
        """
        return TagHelper.SUPPORTED_FORMATS

    @staticmethod
    def generate_tag_object(song: Song) -> Any:
        """
        Generates the object that allows to handle the tags embedded in the audio file.
        :param song: An instance of the class "Song" representing the audio file to process.
        :type song: Song
        :return: An object representing the song's tags.
        :rtype: Any
        :raise ValueError: If an unsupported file type has been defined in the song object.
        """
        extension: str = song.get_extension()
        path: str = song.get_path()
        if not extension or not path:
            return None
        if extension == 'm4a':
            # Generate the object to process some MPEG based files such as Apple ALAC.
            return MP4(path)
        if extension == 'mp3':
            # Generate the object to process MP3 and similar formats.
            return ID3(path)
        elif extension == 'flac':
            # Generate the object to process FLAC encoded files.
            return FLAC(path)
        elif extension == 'aif' or extension == 'aiff':
            # Generate the object to process AIFF encoded files.
            return AIFF(path)
        elif extension == 'ogg':
            # Generate the object to process OGG files.
            return mutagen.File(path)
        else:
            raise ValueError('Unsupported file type.')

    def set_song(self, song: Song) -> None:
        """
        Sets the song that will be processed.
        :param song: An instance of the class "Song" representing the song to process.
        :type song: Song
        """
        self.song = song

    def get_song(self) -> Song:
        """
        Returns the song that will be processed.
        :return: An instance of the class "Song" representing the song to process.
        :rtype Song
        """
        return self.song

    def __init__(self, song: Song):
        """
        The class constructor.
        :param song: An instance of the class "Song" representing the audio file to process.
        :type song: Song
        """
        self.set_song(song)

    def fetch(self) -> None:
        """
        Loads the tags from the audio file defined and then set them into the song object.
        :raise ValueError: If no song has been defined.
        :raise ValueError: If an unsupported file type has been defined in the song object.
        """
        if self.song is None:
            raise ValueError('No song has been defined.')
        tags: Any = self.song.get_tag_object()
        extension: str = self.song.get_extension()
        # Load the song title and artists required to build the search query used by iTunes API and lyrics look up.
        if extension == 'm4a':
            if '©nam' in tags and len(tags['©nam']) > 0:
                self.song.set_title(tags['©nam'][0])
            if '©ART' in tags and len(tags['©ART']) > 0:
                self.song.set_artist(tags['©ART'][0])
        elif extension == 'mp3':
            if 'TIT2' in tags:
                self.song.set_title(str(tags['TIT2']))
            if 'TPE1' in tags:
                self.song.set_artist(str(tags['TPE1']))
        elif extension == 'flac':
            if 'title' in tags and len(tags['title']) > 0:
                self.song.set_title(tags['title'][0])
            if 'artist' in tags and len(tags['artist']) > 0:
                self.song.set_artist(tags['artist'][0])
        elif extension == 'aif' or extension == 'aiff':
            if 'TIT2' in tags:
                self.song.set_title(str(tags['TIT2']))
            if 'TPE1' in tags:
                self.song.set_artist(str(tags['TPE1']))
        elif extension == 'ogg':
            if 'title' in tags and len(tags['title']) > 0:
                self.song.set_title(tags['title'][0])
            if 'artist' in tags and len(tags['artist']) > 0:
                self.song.set_artist(tags['artist'][0])
        else:
            raise ValueError('Unsupported file type.')

    def save(self) -> None:
        """
        Sets the file tags according to song properties.
        """
        if self.song is None:
            raise ValueError('No song has been defined.')
        extension: str = self.song.get_extension()
        if extension == 'm4a':
            self.__save_m4a()
        elif extension == 'mp3':
            self.__save_id3()
        elif extension == 'flac':
            self.__save_flac()
        elif extension == 'aif' or extension == 'aiff':
            self.__save_aiff()
        elif extension == 'ogg':
            self.__save_ogg()
        else:
            raise ValueError('Unsupported file type.')



from diesis import Config
from typing import Optional
from datetime import datetime


class Logger:
    __log_fp = None

    @staticmethod
    def log(message: str) -> None:
        """
        Displays a message to the user console.
        :param message: A string containing the message to show.
        :type message: str
        """
        if message and Config.Config.is_verbose():
            print(message)
        log_file: Optional[str] = Config.Config.get_log_file()
        if message and log_file:
            if not Logger.__log_fp:
                # Open the log file if no file pointer has been found.
                Logger.__log_fp = open(log_file, 'a')
            date: str = datetime.today().strftime('%Y/%m/%d - %H:%M:%S')
            message = '[' + date + ']: ' + message + '\n'
            # Write log message to file including current date.
            Logger.__log_fp.write(message)

    @staticmethod
    def log_error(message: str) -> None:
        """
        Displays an error message to the user console, error messages are yellow colored.
        :param message: A string containing the error message to show.
        :type message: str
        """
        if message and Config.Config.is_verbose():
            print('\033[93m' + message + '\033[0m')



from typing import Optional


class LyricsScraper:
    query: str = None
    minimal_query: str = None
    lyrics: str = None
    lyrics_writer: str = None

    def set_query(self, query: str, minimal_query: str) -> None:
        """
        Sets the search query to use for lyrics look up.
        :param query: A string containing the search query for this song.
        :type query: str
        :param minimal_query: A string containing a shorter query, usually it doesn't include texts inside parenthesis.
        :type minimal_query: str
        """
        self.query = query
        self.minimal_query = minimal_query

    def get_query(self, minimal: bool = False) -> str:
        """
        Returns the search query that will be used for lyrics look up.
        :param minimal: If set to "True", the shorter query version will be returned instead.
        :type minimal: bool
        :return: A string containing the search query, if no query has been defined, an empty string will be returned.
        :rtype: str
        """
        if minimal:
            query: str = self.minimal_query
        else:
            query: str = self.query
        if query is None:
            return ''
        return query

    def get_lyrics(self) -> Optional[str]:
        """
        Returns the lyrics that has been found, before using this method, you must invoke the "fetch" method.
        :return: A string containing the song lyrics found, if no lyrics were found, None will be returned.
        :rtype: Optional[str]
        """
        return self.lyrics

    def get_lyrics_writer(self) -> Optional[str]:
        """
        Returns the authors who wrote the lyrics separated by a comma.
        :return: A string containing the lyrics authors, if no author was found, None will be returned instead.
        :rtype: Optional[str]
        """
        return self.lyrics_writer



from diesis import FileScanner, Config


def main():
    try:
        Config.Config.setup_from_cli()
        print('Running Diesis 1.0.5')
        scanner = FileScanner.FileScanner()
        scanner.scan()
        print('All operations completed, bye!')
    except KeyboardInterrupt:
        print('Execution interrupted, bye!')



import setuptools


with open('README.md', 'r') as description:
    long_description = description.read()

setuptools.setup(
    name='Diesis',
    version='1.0.5',
    license='MIT',
    description='A simple auto-tagging tool for precise music collectors.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Enrico Sola',
    author_email='info@enricosola.com',
    url='https://github.com/RyanJ93/diesis',
    packages=setuptools.find_packages(),
    keywords=['music', 'tagging', 'id3', 'mp3', 'm4a', 'flac', 'ogg'],
    install_requires=[
        'mutagen',
        'beautifulsoup4',
        'pydub'
    ],
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    entry_points={
        'console_scripts': ['diesis=diesis.main:main']
    }
)



from typing import Set, List, Optional
from diesis import Song, Config
from pydub import AudioSegment


class Converter:
    SUPPORTED_FORMATS: Set[str] = {'m4a', 'mp3', 'alac', 'flac', 'aiff', 'aif', 'ogg'}

    song: Song.Song = None

    @staticmethod
    def get_supported_formats() -> Set[str]:
        """
        Returns the extensions of all the supported file types.
        :return: A set containing the supported file extensions.
        :rtype: Set[str]
        """
        return Converter.SUPPORTED_FORMATS

    def __init__(self, song: Song.Song):
        """
        The class constructor.
        :param song: An instance fo the class "Song" representing the song to convert.
        :type song: Song.Song
        """
        self.set_song(song)

    def set_song(self, song: Song.Song) -> None:
        """
        Sets the song to convert.
        :param song: An instance fo the class "Song" representing the song to convert.
        :type song: Song.Song
        """
        self.song = song

    def get_song(self) -> Song.Song:
        """
        Returns the song to convert.
        :return: An instance fo the class "Song" representing the song to convert.
        :rtype: Song.Song
        """
        return self.song

    def convert(self, conversion_format: str) -> str:
        """
        Convert the file contained within the given song object.
        :param conversion_format: A string containing the format the song must be converted into.
        :type conversion_format: str
        :return: A string containing the path to the converted file.
        :rtype: str
        :raise ValueError: If no song has been defined.
        :raise ValueError: If an invalid or an unsupported format is given.
        """
        # TODO: File conversion into Apple ALAC doesn't work due to a ffmpeg error.
        if self.song is None:
            raise ValueError('No song has been defined.')
        if not conversion_format or conversion_format not in Converter.SUPPORTED_FORMATS:
            raise ValueError('Unsupported format.')
        path: str = self.song.get_path()
        extension: str = self.song.get_extension()
        if extension == conversion_format:
            return path
        arguments: List[str] = []
        new_extension: str = conversion_format
        if conversion_format == 'alac':
            # Patches in CLI arguments based on this command sample:
            # ffmpeg -i input.mp3 -vn -c:a alac output.m4a
            # arguments = ['-vn', '-c:a', 'alac']
            arguments = ['-acodec', 'alac']
            conversion_format = 'aac'
            new_extension = 'm4a'
        elif conversion_format == 'aiff' or conversion_format == 'aif':
            extension = 'aac'
        index: int = path.rfind('.') + 1
        new_path: str = path[:index] + new_extension
        # Generate the object for conversion.
        audio: AudioSegment = AudioSegment.from_file(path, extension)
        # Convert and save the file.
        bitrate: Optional[int] = Config.Config.get_bitrate()
        if bitrate and bitrate > 0:
            audio.export(new_path, format=conversion_format, parameters=arguments, bitrate=bitrate)
        else:
            audio.export(new_path, format=conversion_format, parameters=arguments)
        return new_path
